<?xml version="1.0" encoding="UTF-8"?>
<nodes>
<node
    service_host="host"
    cpus="8"
    memory_mb="8192"
    local_gb="128"
    pm_address="10.1.2.3"
    pm_user="pm_user"
    terminal_port="8000"
    instance_uuid="None"
    id="%(node_id)s">
    <interfaces>
         <interface id="%(interface_id)s" address="%(address)s" datapath_id="None" port_no="None"/>
    </interfaces>
</node>
</nodes>
