<?xml version='1.0' encoding='UTF-8'?>
<node instance_uuid="None" pm_address="10.1.2.3" task_state="None" uuid="%(node_uuid)s" pxe_config_path="None" cpus="8" updated_at="None" memory_mb="8192" service_host="host" local_gb="128" id="%(node_id)s" pm_user="pm_user" terminal_port="8000">
  <interfaces>
    <interface datapath_id="None" id="%(interface_id)s" port_no="None" address="%(address)s"/>
  </interfaces>
</node>
