<?xml version='1.0' encoding='UTF-8'?>
<cell xmlns="http://docs.rackspacecloud.com/servers/api/v1.0">
    <capacities>
        <ram_free total_mb="7680">
            <unit_by_mb unit="0"  mb="8192"/>
            <unit_by_mb unit="13" mb="512"/>
            <unit_by_mb unit="1"  mb="4096"/>
            <unit_by_mb unit="3"  mb="2048"/>
            <unit_by_mb unit="0"  mb="16384"/>
        </ram_free>
        <disk_free total_mb="1052672">
            <unit_by_mb unit="11" mb="81920"/>
            <unit_by_mb unit="46" mb="20480"/>
            <unit_by_mb unit="23" mb="40960"/>
            <unit_by_mb unit="5"  mb="163840"/>
            <unit_by_mb unit="0"  mb="0"/>
        </disk_free>
    </capacities>
</cell>
