{
    "flavors": [
        {
            "disabled": false,
            "disk": 1,
            "ephemeral": 0,
            "id": "1",
            "links": [
                {
                    "href": "http://openstack.example.com/v3/flavors/1",
                    "rel": "self"
                },
                {
                    "href": "http://openstack.example.com/flavors/1",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.tiny",
            "flavor-access:is_public": true,
            "ram": 512,
            "os-flavor-rxtx:rxtx_factor": 1.0,
            "swap": 0,
            "vcpus": 1
        },
        {
            "disabled": false,
            "disk": 20,
            "ephemeral": 0,
            "id": "2",
            "links": [
                {
                    "href": "http://openstack.example.com/v3/flavors/2",
                    "rel": "self"
                },
                {
                    "href": "http://openstack.example.com/flavors/2",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.small",
            "flavor-access:is_public": true,
            "ram": 2048,
            "os-flavor-rxtx:rxtx_factor": 1.0,
            "swap": 0,
            "vcpus": 1
        },
        {
            "disabled": false,
            "disk": 40,
            "ephemeral": 0,
            "id": "3",
            "links": [
                {
                    "href": "http://openstack.example.com/v3/flavors/3",
                    "rel": "self"
                },
                {
                    "href": "http://openstack.example.com/flavors/3",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.medium",
            "flavor-access:is_public": true,
            "ram": 4096,
            "os-flavor-rxtx:rxtx_factor": 1.0,
            "swap": 0,
            "vcpus": 2
        },
        {
            "disabled": false,
            "disk": 80,
            "ephemeral": 0,
            "id": "4",
            "links": [
                {
                    "href": "http://openstack.example.com/v3/flavors/4",
                    "rel": "self"
                },
                {
                    "href": "http://openstack.example.com/flavors/4",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.large",
            "flavor-access:is_public": true,
            "ram": 8192,
            "os-flavor-rxtx:rxtx_factor": 1.0,
            "swap": 0,
            "vcpus": 4
        },
        {
            "disabled": false,
            "disk": 160,
            "ephemeral": 0,
            "id": "5",
            "links": [
                {
                    "href": "http://openstack.example.com/v3/flavors/5",
                    "rel": "self"
                },
                {
                    "href": "http://openstack.example.com/flavors/5",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.xlarge",
            "flavor-access:is_public": true,
            "ram": 16384,
            "os-flavor-rxtx:rxtx_factor": 1.0,
            "swap": 0,
            "vcpus": 8
        }
    ]
}
