{
    "snapshot": {
        "volume_id": "%(volume_id)s",
        "create_info": {
            "snapshot_id": "%(snapshot_id)s",
            "type": "%(type)s",
            "new_file": "%(new_file)s"}
     }
}
