<?xml version='1.0' encoding='UTF-8'?>
<events>
  <event>
    <status>%(status)s</status>
    <tag>%(tag)s</tag>
    <name>%(name)s</name>
    <server_uuid>%(uuid)s</server_uuid>
    <code>200</code>
  </event>
</events>
