<?xml version='1.0' encoding='UTF-8'?>
<volumes>
  <volume status="in-use" displayDescription="%(volume_desc)s" availabilityZone="zone1:host1" displayName="%(volume_name)s" volumeType="Backup" snapshotId="None" id="%(uuid)s" createdAt="%(timestamp)s" size="100">
    <attachments>
      <attachment device="/" serverId="%(uuid)s" id="%(uuid)s" volumeId="%(uuid)s"/>
    </attachments>
    <metadata/>
  </volume>
</volumes>
