<?xml version='1.0' encoding='UTF-8'?>
<tenant_usages>
  <tenant_usage>
    <tenant_id>openstack</tenant_id>
    <total_local_gb_usage>1.0</total_local_gb_usage>
    <total_vcpus_usage>1.0</total_vcpus_usage>
    <total_memory_mb_usage>512.0</total_memory_mb_usage>
    <total_hours>1.0</total_hours>
    <start>%(timestamp)s</start>
    <stop>%(timestamp)s</stop>
    <server_usages/>
  </tenant_usage>
</tenant_usages>
