{
    "extensions": [
        {
            "alias": "NMN",
            "description": "%(text)s",
            "links": [],
            "name": "Multinic",
            "namespace": "http://docs.openstack.org/compute/ext/multinic/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-DCF",
            "description": "%(text)s",
            "links": [],
            "name": "DiskConfig",
            "namespace": "http://docs.openstack.org/compute/ext/disk_config/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-EXT-AZ",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedAvailabilityZone",
            "namespace": "http://docs.openstack.org/compute/ext/extended_availability_zone/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-EXT-IPS",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedIps",
            "namespace": "http://docs.openstack.org/compute/ext/extended_ips/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-EXT-IPS-MAC",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedIpsMac",
            "namespace": "http://docs.openstack.org/compute/ext/extended_ips_mac/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-EXT-IMG-SIZE",
            "description": "%(text)s",
            "links": [],
            "name": "ImageSize",
            "namespace": "http://docs.openstack.org/compute/ext/image_size/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-EXT-SRV-ATTR",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedServerAttributes",
            "namespace": "http://docs.openstack.org/compute/ext/extended_status/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-EXT-STS",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedStatus",
            "namespace": "http://docs.openstack.org/compute/ext/extended_status/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-extended-volumes",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedVolumes",
            "namespace": "http://docs.openstack.org/compute/ext/extended_volumes/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-EXT-VIF-NET",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedVIFNet",
            "namespace": "http://docs.openstack.org/compute/ext/extended-virtual-interfaces-net/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-FLV-DISABLED",
            "description": "%(text)s",
            "links": [],
            "name": "FlavorDisabled",
            "namespace": "http://docs.openstack.org/compute/ext/flavor_disabled/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-FLV-EXT-DATA",
            "description": "%(text)s",
            "links": [],
            "name": "FlavorExtraData",
            "namespace": "http://docs.openstack.org/compute/ext/flavor_extra_data/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-SRV-USG",
            "description": "%(text)s",
            "links": [],
            "name": "ServerUsage",
            "namespace": "http://docs.openstack.org/compute/ext/server_usage/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-console-auth-tokens",
            "description": "%(text)s",
            "links": [],
            "name": "ConsoleAuthTokens",
            "namespace": "http://docs.openstack.org/compute/ext/consoles-auth-tokens/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "OS-SCH-HNT",
            "description": "%(text)s",
            "links": [],
            "name": "SchedulerHints",
            "namespace": "http://docs.openstack.org/compute/ext/scheduler-hints/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-admin-actions",
            "description": "%(text)s",
            "links": [],
            "name": "AdminActions",
            "namespace": "http://docs.openstack.org/compute/ext/admin-actions/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-aggregates",
            "description": "%(text)s",
            "links": [],
            "name": "Aggregates",
            "namespace": "http://docs.openstack.org/compute/ext/aggregates/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-agents",
            "description": "%(text)s",
            "links": [],
            "name": "Agents",
            "namespace": "http://docs.openstack.org/compute/ext/agents/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-assisted-volume-snapshots",
            "description": "%(text)s",
            "links": [],
            "name": "AssistedVolumeSnapshots",
            "namespace": "http://docs.openstack.org/compute/ext/assisted-volume-snapshots/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-attach-interfaces",
            "description": "Attach interface support.",
            "links": [],
            "name": "AttachInterfaces",
            "namespace": "http://docs.openstack.org/compute/ext/interfaces/api/v1.1",
            "updated": "2012-07-22T00:00:00+00:00"
        },
        {
            "alias": "os-availability-zone",
            "description": "%(text)s",
            "links": [],
            "name": "AvailabilityZone",
            "namespace": "http://docs.openstack.org/compute/ext/availabilityzone/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-baremetal-nodes",
            "description": "%(text)s",
            "links": [],
            "name": "BareMetalNodes",
            "namespace": "http://docs.openstack.org/compute/ext/baremetal_nodes/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-block-device-mapping-v2-boot",
            "description": "%(text)s",
            "links": [],
            "name": "BlockDeviceMappingV2Boot",
            "namespace": "http://docs.openstack.org/compute/ext/block_device_mapping_v2_boot/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-cells",
            "description": "%(text)s",
            "links": [],
            "name": "Cells",
            "namespace": "http://docs.openstack.org/compute/ext/cells/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-cell-capacities",
            "description": "%(text)s",
            "links": [],
            "name": "CellCapacities",
            "namespace": "http://docs.openstack.org/compute/ext/cell_capacities/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-certificates",
            "description": "%(text)s",
            "links": [],
            "name": "Certificates",
            "namespace": "http://docs.openstack.org/compute/ext/certificates/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-cloudpipe",
            "description": "%(text)s",
            "links": [],
            "name": "Cloudpipe",
            "namespace": "http://docs.openstack.org/compute/ext/cloudpipe/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-cloudpipe-update",
            "description": "%(text)s",
            "links": [],
            "name": "CloudpipeUpdate",
            "namespace": "http://docs.openstack.org/compute/ext/cloudpipe-update/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-config-drive",
            "description": "%(text)s",
            "links": [],
            "name": "ConfigDrive",
            "namespace": "http://docs.openstack.org/compute/ext/config_drive/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-console-output",
            "description": "%(text)s",
            "links": [],
            "name": "ConsoleOutput",
            "namespace": "http://docs.openstack.org/compute/ext/os-console-output/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-consoles",
            "description": "%(text)s",
            "links": [],
            "name": "Consoles",
            "namespace": "http://docs.openstack.org/compute/ext/os-consoles/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-create-server-ext",
            "description": "%(text)s",
            "links": [],
            "name": "Createserverext",
            "namespace": "http://docs.openstack.org/compute/ext/createserverext/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-deferred-delete",
            "description": "%(text)s",
            "links": [],
            "name": "DeferredDelete",
            "namespace": "http://docs.openstack.org/compute/ext/deferred-delete/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-evacuate",
            "description": "%(text)s",
            "links": [],
            "name": "Evacuate",
            "namespace": "http://docs.openstack.org/compute/ext/evacuate/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-extended-floating-ips",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedFloatingIps",
            "namespace": "http://docs.openstack.org/compute/ext/extended_floating_ips/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-fixed-ips",
            "description": "Fixed IPs support.",
            "links": [],
            "name": "FixedIPs",
            "namespace": "http://docs.openstack.org/compute/ext/fixed_ips/api/v2",
            "updated": "2012-10-18T13:25:27-06:00"
        },
        {
            "alias": "os-flavor-access",
            "description": "%(text)s",
            "links": [],
            "name": "FlavorAccess",
            "namespace": "http://docs.openstack.org/compute/ext/flavor_access/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-flavor-extra-specs",
            "description": "%(text)s",
            "links": [],
            "name": "FlavorExtraSpecs",
            "namespace": "http://docs.openstack.org/compute/ext/flavor_extra_specs/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-flavor-manage",
            "description": "%(text)s",
            "links": [],
            "name": "FlavorManage",
            "namespace": "http://docs.openstack.org/compute/ext/flavor_manage/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-flavor-rxtx",
            "description": "%(text)s",
            "links": [],
            "name": "FlavorRxtx",
            "namespace": "http://docs.openstack.org/compute/ext/flavor_rxtx/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-flavor-swap",
            "description": "%(text)s",
            "links": [],
            "name": "FlavorSwap",
            "namespace": "http://docs.openstack.org/compute/ext/flavor_swap/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-floating-ip-dns",
            "description": "%(text)s",
            "links": [],
            "name": "FloatingIpDns",
            "namespace": "http://docs.openstack.org/ext/floating_ip_dns/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-floating-ip-pools",
            "description": "%(text)s",
            "links": [],
            "name": "FloatingIpPools",
            "namespace": "http://docs.openstack.org/compute/ext/floating_ip_pools/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-floating-ips",
            "description": "%(text)s",
            "links": [],
            "name": "FloatingIps",
            "namespace": "http://docs.openstack.org/compute/ext/floating_ips/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-floating-ips-bulk",
            "description": "%(text)s",
            "links": [],
            "name": "FloatingIpsBulk",
            "namespace": "http://docs.openstack.org/compute/ext/floating_ips_bulk/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-hide-server-addresses",
            "description": "Support hiding server addresses in certain states.",
            "links": [],
            "name": "HideServerAddresses",
            "namespace": "http://docs.openstack.org/compute/ext/hide_server_addresses/api/v1.1",
            "updated": "2012-12-11T00:00:00+00:00"
        },
        {
            "alias": "os-hosts",
            "description": "%(text)s",
            "links": [],
            "name": "Hosts",
            "namespace": "http://docs.openstack.org/compute/ext/hosts/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-services",
            "description": "%(text)s",
            "links": [],
            "name": "Services",
            "namespace": "http://docs.openstack.org/compute/ext/services/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-extended-services",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedServices",
            "namespace": "http://docs.openstack.org/compute/ext/extended_services/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-fping",
            "description": "%(text)s",
            "links": [],
            "name": "Fping",
            "namespace": "http://docs.openstack.org/compute/ext/fping/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-hypervisors",
            "description": "%(text)s",
            "links": [],
            "name": "Hypervisors",
            "namespace": "http://docs.openstack.org/compute/ext/hypervisors/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-extended-hypervisors",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedHypervisors",
            "namespace": "http://docs.openstack.org/compute/ext/extended_hypervisors/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-server-external-events",
            "description": "%(text)s",
            "links": [],
            "name": "ServerExternalEvents",
            "namespace": "http://docs.openstack.org/compute/ext/server-external-events/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-instance_usage_audit_log",
            "description": "%(text)s",
            "links": [],
            "name": "OSInstanceUsageAuditLog",
            "namespace": "http://docs.openstack.org/ext/services/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-keypairs",
            "description": "%(text)s",
            "links": [],
            "name": "Keypairs",
            "namespace": "http://docs.openstack.org/compute/ext/keypairs/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-multiple-create",
            "description": "%(text)s",
            "links": [],
            "name": "MultipleCreate",
            "namespace": "http://docs.openstack.org/compute/ext/multiplecreate/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-networks",
            "description": "%(text)s",
            "links": [],
            "name": "Networks",
            "namespace": "http://docs.openstack.org/compute/ext/os-networks/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-tenant-networks",
            "description": "%(text)s",
            "links": [],
            "name": "OSTenantNetworks",
            "namespace": "http://docs.openstack.org/compute/ext/os-tenant-networks/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-networks-associate",
            "description": "%(text)s",
            "links": [],
            "name": "NetworkAssociationSupport",
            "namespace": "http://docs.openstack.org/compute/ext/networks_associate/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-extended-quotas",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedQuotas",
            "namespace": "http://docs.openstack.org/compute/ext/extended_quotas/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-quota-sets",
            "description": "%(text)s",
            "links": [],
            "name": "Quotas",
            "namespace": "http://docs.openstack.org/compute/ext/quotas-sets/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-user-quotas",
            "description": "%(text)s",
            "links": [],
            "name": "UserQuotas",
            "namespace": "http://docs.openstack.org/compute/ext/user_quotas/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-rescue",
            "description": "%(text)s",
            "links": [],
            "name": "Rescue",
            "namespace": "http://docs.openstack.org/compute/ext/rescue/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-security-group-default-rules",
            "description": "%(text)s",
            "links": [],
            "name": "SecurityGroupDefaultRules",
            "namespace": "http://docs.openstack.org/compute/ext/securitygroupdefaultrules/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-security-groups",
            "description": "%(text)s",
            "links": [],
            "name": "SecurityGroups",
            "namespace": "http://docs.openstack.org/compute/ext/securitygroups/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-server-diagnostics",
            "description": "%(text)s",
            "links": [],
            "name": "ServerDiagnostics",
            "namespace": "http://docs.openstack.org/compute/ext/server-diagnostics/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-server-password",
            "description": "%(text)s",
            "links": [],
            "name": "ServerPassword",
            "namespace": "http://docs.openstack.org/compute/ext/server-password/api/v2",
            "updated": "2012-11-29T00:00:00+00:00"
        },
        {
            "alias": "os-server-start-stop",
            "description": "%(text)s",
            "links": [],
            "name": "ServerStartStop",
            "namespace": "http://docs.openstack.org/compute/ext/servers/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-shelve",
            "description": "%(text)s",
            "links": [],
            "name": "Shelve",
            "namespace": "http://docs.openstack.org/compute/ext/shelve/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-simple-tenant-usage",
            "description": "%(text)s",
            "links": [],
            "name": "SimpleTenantUsage",
            "namespace": "http://docs.openstack.org/compute/ext/os-simple-tenant-usage/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-used-limits",
            "description": "%(text)s",
            "links": [],
            "name": "UsedLimits",
            "namespace": "http://docs.openstack.org/compute/ext/used_limits/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-used-limits-for-admin",
            "description": "%(text)s",
            "links": [],
            "name": "UsedLimitsForAdmin",
            "namespace": "http://docs.openstack.org/compute/ext/used_limits_for_admin/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-user-data",
            "description": "%(text)s",
            "links": [],
            "name": "UserData",
            "namespace": "http://docs.openstack.org/compute/ext/userdata/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-virtual-interfaces",
            "description": "%(text)s",
            "links": [],
            "name": "VirtualInterfaces",
            "namespace": "http://docs.openstack.org/compute/ext/virtual_interfaces/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-volume-attachment-update",
            "description": "%(text)s",
            "links": [],
            "name": "VolumeAttachmentUpdate",
            "namespace": "http://docs.openstack.org/compute/ext/os-volume-attachment-update/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-volumes",
            "description": "%(text)s",
            "links": [],
            "name": "Volumes",
            "namespace": "http://docs.openstack.org/compute/ext/volumes/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-instance-actions",
            "description": "%(text)s",
            "links": [],
            "name": "InstanceActions",
            "namespace": "http://docs.openstack.org/compute/ext/instance-actions/api/v1.1",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-migrations",
            "description": "%(text)s",
            "links": [],
            "name": "Migrations",
            "namespace": "http://docs.openstack.org/compute/ext/migrations/api/v2.0",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-baremetal-ext-status",
            "description": "%(text)s",
            "links": [],
            "name": "BareMetalExtStatus",
            "namespace": "http://docs.openstack.org/compute/ext/baremetal_ext_status/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-preserve-ephemeral-rebuild",
            "description": "%(text)s",
            "links": [],
            "name": "PreserveEphemeralOnRebuild",
            "namespace": "http://docs.openstack.org/compute/ext/preserve_ephemeral_rebuild/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-extended-services-delete",
            "description": "%(text)s",
            "links": [],
            "name": "ExtendedServicesDelete",
            "namespace": "http://docs.openstack.org/compute/ext/extended_services_delete/api/v2",
            "updated": "%(timestamp)s"
        },
        {
            "alias": "os-server-groups",
            "description": "%(text)s",
            "links": [],
            "name": "ServerGroups",
            "namespace": "http://docs.openstack.org/compute/ext/servergroups/api/v2",
            "updated": "%(timestamp)s"
        }
    ]
}
