<?xml version="1.0" encoding="UTF-8"?>
<interfaceAttachments>
  <interfaceAttachment>
    <port_state>%(port_state)s</port_state>
    <fixed_ips>
      <fixed_ip>
	<subnet_id>%(subnet_id)s</subnet_id>
	<ip_address>%(ip_address)s</ip_address>
      </fixed_ip>
    </fixed_ips>
    <port_id>%(port_id)s</port_id>
    <net_id>%(net_id)s</net_id>
    <mac_addr>%(mac_addr)s</mac_addr>
 </interfaceAttachment>
</interfaceAttachments>
