<?xml version='1.0' encoding='UTF-8'?>
<instance_usage_audit_logs>
  <total_errors>0</total_errors>
  <total_instances>0</total_instances>
  <log/>
  <num_hosts_running>0</num_hosts_running>
  <num_hosts_done>0</num_hosts_done>
  <num_hosts_not_run>1</num_hosts_not_run>
  <hosts_not_run>
    <item>%(hostid)s</item>
  </hosts_not_run>
  <overall_status>0 of 1 hosts done. 0 errors.</overall_status>
  <period_ending>%(timestamp)s</period_ending>
  <period_beginning>%(timestamp)s</period_beginning>
  <num_hosts>1</num_hosts>
</instance_usage_audit_logs>
