<?xml version='1.0' encoding='UTF-8'?>
<keypair>
  <public_key>%(public_key)s</public_key>
  <user_id>fake</user_id>
  <name>%(keypair_name)s</name>
  <fingerprint>%(fingerprint)s</fingerprint>
</keypair>
