<?xml version='1.0' encoding='UTF-8'?>
<console>
    <instance_uuid>%(id)s</instance_uuid>
    <host>%(host)s</host>
    <port>%(port)s</port>
    <internal_access_path>%(internal_access_path)s</internal_access_path>
</console>
