<?xml version='1.0' encoding='UTF-8'?>
<host>
  <resource>
    <project>(total)</project>
    <memory_mb>8192</memory_mb>
    <host>%(host_name)s</host>
    <cpu>1</cpu>
    <disk_gb>1028</disk_gb>
  </resource>
  <resource>
    <project>(used_now)</project>
    <memory_mb>512</memory_mb>
    <host>%(host_name)s</host>
    <cpu>0</cpu>
    <disk_gb>0</disk_gb>
  </resource>
  <resource>
    <project>(used_max)</project>
    <memory_mb>0</memory_mb>
    <host>%(host_name)s</host>
    <cpu>0</cpu>
    <disk_gb>0</disk_gb>
  </resource>
</host>
