{
    "node": {
        "service_host": "host",
        "cpus": 8,
        "memory_mb": 8192,
        "local_gb": 128,
        "pm_address": "10.1.2.3",
        "pm_user": "pm_user",
        "pm_password": "pm_pass",
        "terminal_port": 8000
    }
}
