<?xml version='1.0' encoding='UTF-8'?>
<keypairs>
  <keypair>
    <public_key>%(public_key)s
</public_key>
    <name>%(keypair_name)s</name>
    <fingerprint>%(fingerprint)s</fingerprint>
  </keypair>
</keypairs>
