<?xml version='1.0' encoding='UTF-8'?>
<cells xmlns="http://docs.rackspacecloud.com/servers/api/v1.0">
  <cell name="cell1" username="username1" rpc_port="None" rpc_host="None" type="child"/>
  <cell name="cell2" username="username2" rpc_port="None" rpc_host="None" type="parent"/>
  <cell name="cell3" username="username3" rpc_port="None" rpc_host="None" type="child"/>
  <cell name="cell4" username="username4" rpc_port="None" rpc_host="None" type="parent"/>
  <cell name="cell5" username="username5" rpc_port="None" rpc_host="None" type="child"/>
</cells>
