<?xml version='1.0' encoding='UTF-8'?>
<security_group xmlns="http://docs.openstack.org/compute/api/v1.1" tenant_id="openstack" id="2" name="%(group_name)s">
  <description>
        description
    </description>
  <rules/>
</security_group>
