{
    "images": [
        {
            "OS-DCF:diskConfig": "AUTO",
            "created": "%(timestamp)s",
            "id": "%(id)s",
            "links": [
                {
                    "href": "%(host)s/v2/openstack/images/%(id)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/openstack/images/%(id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "%(glance_host)s/openstack/images/%(id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "architecture": "x86_64",
                "auto_disk_config": "True",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage7",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "%(timestamp)s"
        },
        {
            "created": "%(timestamp)s",
            "id": "%(id)s",
            "links": [
                {
                    "href": "%(host)s/v2/openstack/images/%(id)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/openstack/images/%(id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "%(glance_host)s/openstack/images/%(id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "architecture": "x86_64",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "%(timestamp)s"
        },
        {
            "created": "%(timestamp)s",
            "id": "%(uuid)s",
            "links": [
                {
                    "href": "%(host)s/v2/openstack/images/%(uuid)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/openstack/images/%(uuid)s",
                    "rel": "bookmark"
                },
                {
                    "href": "%(glance_host)s/openstack/images/%(uuid)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "%(timestamp)s"
        },
        {
            "OS-DCF:diskConfig": "MANUAL",
            "created": "%(timestamp)s",
            "id": "%(uuid)s",
            "links": [
                {
                    "href": "%(host)s/v2/openstack/images/%(uuid)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/openstack/images/%(uuid)s",
                    "rel": "bookmark"
                },
                {
                    "href": "%(glance_host)s/openstack/images/%(uuid)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "architecture": "x86_64",
                "auto_disk_config": "False",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage6",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "%(timestamp)s"
        },
        {
            "created": "%(timestamp)s",
            "id": "%(id)s",
            "links": [
                {
                    "href": "%(host)s/v2/openstack/images/%(id)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/openstack/images/%(id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "%(glance_host)s/openstack/images/%(id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "kernel_id": "%(id)s",
                "ramdisk_id": null
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "%(timestamp)s"
        },
        {
            "created": "%(timestamp)s",
            "id": "%(id)s",
            "links": [
                {
                    "href": "%(host)s/v2/openstack/images/%(id)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/openstack/images/%(id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "%(glance_host)s/openstack/images/%(id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "%(timestamp)s"
        },
        {
            "created": "%(timestamp)s",
            "id": "%(id)s",
            "links": [
                {
                    "href": "%(host)s/v2/openstack/images/%(id)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/openstack/images/%(id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "%(glance_host)s/openstack/images/%(id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "%(timestamp)s"
        }
    ]
}
