<?xml version="1.0" encoding="UTF-8" ?>
    <os-migrateLive>
        <host>%(hostname)s</host>
        <block_migration>false</block_migration>
        <disk_over_commit>false</disk_over_commit>
    </os-migrateLive>
