<?xml version="1.0" encoding="UTF-8"?>
<resize xmlns:OS-DCF="http://docs.openstack.org/compute/ext/disk_config/api/v1.1"
        xmlns:atom="http://www.w3.org/2005/Atom"
        xmlns="http://docs.openstack.org/compute/api/v1.1"
        OS-DCF:diskConfig="AUTO"
        flavorRef="3" />
