<?xml version="1.0" encoding="UTF-8"?>
  <snapshot>
    <volume_id>%(volume_id)s</volume_id>
    <create_info>
      <snapshot_id>%(snapshot_id)s</snapshot_id>
      <type>%(type)s</type>
      <new_file>%(new_file)s</new_file>
    </create_info>
  </snapshot>
