<?xml version="1.0" encoding="UTF-8"?>
<node
    service_host="host"
    cpus="8"
    memory_mb="8192"
    local_gb="128"
    pm_address="10.1.2.3"
    pm_user="pm_user"
    prov_mac_address="%(address)s"
    terminal_port="8000"
/>
