<network>
    <bridge>None</bridge>
    <vpn_public_port>None</vpn_public_port>
    <dhcp_start>%(ip)s</dhcp_start>
    <bridge_interface>None</bridge_interface>
    <updated_at>None</updated_at>
    <id>%(id)s</id>
    <cidr_v6>None</cidr_v6>
    <deleted_at>None</deleted_at>
    <gateway>%(ip)s</gateway>
    <rxtx_base>None</rxtx_base>
    <label>new net 111</label>
    <priority>None</priority>
    <project_id>None</project_id>
    <vpn_private_address>None</vpn_private_address>
    <deleted>False</deleted>
    <vlan>None</vlan>
    <broadcast>%(ip)s</broadcast>
    <netmask>%(ip)s</netmask>
    <injected>None</injected>
    <cidr>10.20.105.0/24</cidr>
    <vpn_public_address>None</vpn_public_address>
    <multi_host>None</multi_host>
    <dns2>None</dns2>
    <created_at>None</created_at>
    <host>None</host>
    <gateway_v6>None</gateway_v6>
    <netmask_v6>None</netmask_v6>
    <dns1>None</dns1>
</network>
