<?xml version='1.0' encoding='UTF-8'?>
<diagnostics xmlns="http://docs.openstack.org/compute/api/v1.1">
  <memory>524288</memory>
  <vda_write>5778432</vda_write>
  <vnet1_rx_packets>26701</vnet1_rx_packets>
  <vnet1_rx_drop>0</vnet1_rx_drop>
  <vnet1_tx_drop>0</vnet1_tx_drop>
  <vnet1_rx>2070139</vnet1_rx>
  <vda_read>262144</vda_read>
  <vda_write_req>488</vda_write_req>
  <vnet1_tx>140208</vnet1_tx>
  <vnet1_tx_errors>0</vnet1_tx_errors>
  <vnet1_tx_packets>662</vnet1_tx_packets>
  <vnet1_rx_errors>0</vnet1_rx_errors>
  <cpu0_time>17300000000</cpu0_time>
  <vda_read_req>112</vda_read_req>
  <vda_errors>-1</vda_errors>
</diagnostics>