<?xml version="1.0" encoding="UTF-8"?>
<networks>
  <network>
    <cidr>10.0.0.0/29</cidr>
    <id>%(id)s</id>
    <label>test_0</label>
  </network>
  <network>
    <cidr>10.0.0.8/29</cidr>
    <id>%(id)s</id>
    <label>test_1</label>
  </network>
</networks>
