<?xml version='1.0' encoding='UTF-8'?>
<availabilityZones
    xmlns:os-availability-zone="http://docs.openstack.org/compute/ext/availabilityzone/api/v1.1">
    <availabilityZone name="zone-1">
        <zoneState available="True" />
        <metadata />
    </availabilityZone>
    <availabilityZone name="zone-2">
        <zoneState available="False" />
        <metadata />
    </availabilityZone>
</availabilityZones>