<?xml version="1.0" encoding="UTF-8"?>
<agent>
  <url>%(url)s</url>
  <hypervisor>%(hypervisor)s</hypervisor>
  <md5hash>%(md5hash)s</md5hash>
  <version>%(version)s</version>
  <architecture>%(architecture)s</architecture>
  <os>%(os)s</os>
  <agent_id>%(agent_id)d</agent_id>
</agent>
