{
    "interfaceAttachment": {
        "fixed_ips": [{
	    "subnet_id": "%(subnet_id)s",
            "ip_address": "%(ip_address)s"
        }],
        "mac_addr": "fa:16:3e:4c:2c:30",
        "net_id": "%(net_id)s",
        "port_id": "%(port_id)s",
	"port_state": "%(port_state)s"
    }
}
