{
    "hypervisor": {
        "cpu_info": "?",
        "current_workload": 0,
        "disk_available_least": 0,
        "host_ip": "%(ip)s",
        "free_disk_gb": 1028,
        "free_ram_mb": 7680,
        "hypervisor_hostname": "fake-mini",
        "hypervisor_type": "fake",
        "hypervisor_version": 1,
        "id": 1,
        "local_gb": 1028,
        "local_gb_used": 0,
        "memory_mb": 8192,
        "memory_mb_used": 512,
        "os-pci:pci_stats": [
            {
                "count": 5,
                "extra_info": {
                    "key1": "value1",
                    "phys_function": "[[\"0x0000\", \"0x04\", \"0x00\", \"0x1\"]]"
                },
                "keya": "valuea",
                "product_id": "1520",
                "vendor_id": "8086"
            }
        ],
        "running_vms": 0,
        "service": {
            "host": "043b3cacf6f34c90a7245151fc8ebcda",
            "id": 2
        },
        "vcpus": 1,
        "vcpus_used": 0
    }
}
