<?xml version='1.0' encoding='UTF-8'?>
<flavors xmlns:atom="http://www.w3.org/2005/Atom" xmlns="http://docs.openstack.org/compute/api/v1.1">
  <flavor name="m1.tiny" id="1">
    <atom:link href="http://openstack.example.com/v2/openstack/flavors/1" rel="self"/>
    <atom:link href="http://openstack.example.com/openstack/flavors/1" rel="bookmark"/>
  </flavor>
  <flavor name="m1.small" id="2">
    <atom:link href="http://openstack.example.com/v2/openstack/flavors/2" rel="self"/>
    <atom:link href="http://openstack.example.com/openstack/flavors/2" rel="bookmark"/>
  </flavor>
  <flavor name="m1.medium" id="3">
    <atom:link href="http://openstack.example.com/v2/openstack/flavors/3" rel="self"/>
    <atom:link href="http://openstack.example.com/openstack/flavors/3" rel="bookmark"/>
  </flavor>
  <flavor name="m1.large" id="4">
    <atom:link href="http://openstack.example.com/v2/openstack/flavors/4" rel="self"/>
    <atom:link href="http://openstack.example.com/openstack/flavors/4" rel="bookmark"/>
  </flavor>
  <flavor name="m1.xlarge" id="5">
    <atom:link href="http://openstack.example.com/v2/openstack/flavors/5" rel="self"/>
    <atom:link href="http://openstack.example.com/openstack/flavors/5" rel="bookmark"/>
  </flavor>
</flavors>