{
    "cells": [
        {
            "name": "cell1",
            "username": "username1",
            "rpc_host": null,
            "rpc_port": null,
            "type": "child"
        },
        {
            "name": "cell2",
            "username": "username2",
            "rpc_host": null,
            "rpc_port": null,
            "type": "parent"
        },
        {
            "name": "cell3",
            "username": "username3",
            "rpc_host": null,
            "rpc_port": null,
            "type": "child"
        },
        {
            "name": "cell4",
            "username": "username4",
            "rpc_host": null,
            "rpc_port": null,
            "type": "parent"
        },
        {
            "name": "cell5",
            "username": "username5",
            "rpc_host": null,
            "rpc_port": null,
            "type": "child"
        }
    ]
}
