<?xml version='1.0' encoding='UTF-8'?>
<aggregate>
  <name>name</name>
  <availability_zone>nova</availability_zone>
  <deleted>False</deleted>
  <created_at>%(timestamp)s</created_at>
  <updated_at>None</updated_at>
  <deleted_at>None</deleted_at>
  <id>%(aggregate_id)s</id>
</aggregate>
