<?xml version='1.0' encoding='UTF-8'?>
<images xmlns:OS-DCF="http://docs.openstack.org/compute/ext/disk_config/api/v1.1" xmlns:atom="http://www.w3.org/2005/Atom" xmlns="http://docs.openstack.org/compute/api/v1.1">
  <image status="ACTIVE" updated="%(timestamp)s" name="fakeimage7" created="%(timestamp)s" minDisk="0" progress="100" minRam="0" id="%(id)s" OS-DCF:diskConfig="AUTO">
    <metadata>
      <meta key="kernel_id">nokernel</meta>
      <meta key="auto_disk_config">True</meta>
      <meta key="ramdisk_id">nokernel</meta>
      <meta key="architecture">x86_64</meta>
    </metadata>
    <atom:link href="%(host)s/v2/openstack/images/%(id)s" rel="self"/>
    <atom:link href="%(host)s/openstack/images/%(id)s" rel="bookmark"/>
    <atom:link href="%(glance_host)s/openstack/images/%(id)s" type="application/vnd.openstack.image" rel="alternate"/>
  </image>
  <image status="ACTIVE" updated="%(timestamp)s" name="fakeimage123456" created="%(timestamp)s" minDisk="0" progress="100" minRam="0" id="%(id)s">
    <metadata>
      <meta key="kernel_id">nokernel</meta>
      <meta key="ramdisk_id">nokernel</meta>
      <meta key="architecture">x86_64</meta>
    </metadata>
    <atom:link href="%(host)s/v2/openstack/images/%(id)s" rel="self"/>
    <atom:link href="%(host)s/openstack/images/%(id)s" rel="bookmark"/>
    <atom:link href="%(glance_host)s/openstack/images/%(id)s" type="application/vnd.openstack.image" rel="alternate"/>
  </image>
  <image status="ACTIVE" updated="%(timestamp)s" name="fakeimage123456" created="%(timestamp)s" minDisk="0" progress="100" minRam="0" id="%(id)s">
    <metadata>
      <meta key="kernel_id">nokernel</meta>
      <meta key="ramdisk_id">nokernel</meta>
    </metadata>
    <atom:link href="%(host)s/v2/openstack/images/%(id)s" rel="self"/>
    <atom:link href="%(host)s/openstack/images/%(id)s" rel="bookmark"/>
    <atom:link href="%(glance_host)s/openstack/images/%(id)s" type="application/vnd.openstack.image" rel="alternate"/>
  </image>
  <image status="ACTIVE" updated="%(timestamp)s" name="fakeimage6" created="%(timestamp)s" minDisk="0" progress="100" minRam="0" id="%(id)s" OS-DCF:diskConfig="MANUAL">
    <metadata>
      <meta key="kernel_id">nokernel</meta>
      <meta key="auto_disk_config">False</meta>
      <meta key="ramdisk_id">nokernel</meta>
      <meta key="architecture">x86_64</meta>
    </metadata>
    <atom:link href="%(host)s/v2/openstack/images/%(id)s" rel="self"/>
    <atom:link href="%(host)s/openstack/images/%(id)s" rel="bookmark"/>
    <atom:link href="%(glance_host)s/openstack/images/%(id)s" type="application/vnd.openstack.image" rel="alternate"/>
  </image>
  <image status="ACTIVE" updated="%(timestamp)s" name="fakeimage123456" created="%(timestamp)s" minDisk="0" progress="100" minRam="0" id="%(id)s">
    <metadata>
      <meta key="kernel_id">%(id)s</meta>
      <meta key="ramdisk_id">None</meta>
    </metadata>
    <atom:link href="%(host)s/v2/openstack/images/%(id)s" rel="self"/>
    <atom:link href="%(host)s/openstack/images/%(id)s" rel="bookmark"/>
    <atom:link href="%(glance_host)s/openstack/images/%(id)s" type="application/vnd.openstack.image" rel="alternate"/>
  </image>
  <image status="ACTIVE" updated="%(timestamp)s" name="fakeimage123456" created="%(timestamp)s" minDisk="0" progress="100" minRam="0" id="%(id)s">
    <metadata>
      <meta key="kernel_id">nokernel</meta>
      <meta key="ramdisk_id">nokernel</meta>
    </metadata>
    <atom:link href="%(host)s/v2/openstack/images/%(id)s" rel="self"/>
    <atom:link href="%(host)s/openstack/images/%(id)s" rel="bookmark"/>
    <atom:link href="%(glance_host)s/openstack/images/%(id)s" type="application/vnd.openstack.image" rel="alternate"/>
  </image>
  <image status="ACTIVE" updated="%(timestamp)s" name="fakeimage123456" created="%(timestamp)s" minDisk="0" progress="100" minRam="0" id="%(id)s">
    <metadata>
      <meta key="kernel_id">nokernel</meta>
      <meta key="ramdisk_id">nokernel</meta>
    </metadata>
    <atom:link href="%(host)s/v2/openstack/images/%(id)s" rel="self"/>
    <atom:link href="%(host)s/openstack/images/%(id)s" rel="bookmark"/>
    <atom:link href="%(glance_host)s/openstack/images/%(id)s" type="application/vnd.openstack.image" rel="alternate"/>
  </image>
</images>
