{
    "instance_usage_audit_logs": {
        "hosts_not_run": [
            "%(hostid)s"
        ],
        "log": {},
        "num_hosts": 1,
        "num_hosts_done": 0,
        "num_hosts_not_run": 1,
        "num_hosts_running": 0,
        "overall_status": "0 of 1 hosts done. 0 errors.",
        "period_beginning": "%(timestamp)s",
        "period_ending": "%(timestamp)s",
        "total_errors": 0,
        "total_instances": 0
    }
}
