<?xml version="1.0" encoding="UTF-8"?>
<flavor xmlns="http://docs.openstack.org/compute/api/v1.1"
    xmlns:OS-FLV-EXT-DATA="http://docs.openstack.org/compute/ext/flavor_extra_data/api/v1.1"
    name="%(flavor_name)s"
    ram="1024"
    vcpus="2"
    disk="10"
    id="%(flavor_id)s"
    rxtx_factor="2.0" />
