<?xml version='1.0' encoding='UTF-8'?>
<networks>
    <network>
        <bridge>br100</bridge>
        <vpn_public_port>1000</vpn_public_port>
        <dhcp_start>%(ip)s</dhcp_start>
        <bridge_interface>eth0</bridge_interface>
        <updated_at>%(timestamp)s</updated_at>
        <id>%(id)s</id>
        <cidr_v6>None</cidr_v6>
        <deleted_at>None</deleted_at>
        <gateway>%(ip)s</gateway>
        <rxtx_base>None</rxtx_base>
        <label>mynet_0</label>
        <priority>None</priority>
        <project_id>1234</project_id>
        <vpn_private_address>%(ip)s</vpn_private_address>
        <deleted>False</deleted>
        <vlan>100</vlan>
        <broadcast>%(ip)s</broadcast>
        <netmask>%(ip)s</netmask>
        <injected>False</injected>
        <cidr>10.0.0.0/29</cidr>
        <vpn_public_address>%(ip)s</vpn_public_address>
        <multi_host>False</multi_host>
        <dns2>None</dns2>
        <created_at>%(timestamp)s</created_at>
        <host>nsokolov-desktop</host>
        <gateway_v6>None</gateway_v6>
        <netmask_v6>None</netmask_v6>
        <dns1>None</dns1>
    </network>
    <network>
        <bridge>br101</bridge>
        <vpn_public_port>1001</vpn_public_port>
        <dhcp_start>%(ip)s</dhcp_start>
        <bridge_interface>eth0</bridge_interface>
        <updated_at>None</updated_at>
        <id>%(id)s</id>
        <cidr_v6>None</cidr_v6>
        <deleted_at>None</deleted_at>
        <gateway>%(ip)s</gateway>
        <rxtx_base>None</rxtx_base>
        <label>mynet_1</label>
        <priority>None</priority>
        <project_id>None</project_id>
        <vpn_private_address>%(ip)s</vpn_private_address>
        <deleted>False</deleted>
        <vlan>101</vlan>
        <broadcast>%(ip)s</broadcast>
        <netmask>%(ip)s</netmask>
        <injected>False</injected>
        <cidr>10.0.0.10/29</cidr>
        <vpn_public_address>None</vpn_public_address>
        <multi_host>False</multi_host>
        <dns2>None</dns2>
        <created_at>%(timestamp)s</created_at>
        <host>None</host>
        <gateway_v6>None</gateway_v6>
        <netmask_v6>None</netmask_v6>
        <dns1>None</dns1>
     </network>
</networks>
