<server
        xmlns="http://docs.openstack.org/compute/api/v1.1"
        xmlns:OS-SCH-HNT="http://docs.openstack.org/compute/ext/scheduler-hints/api/v2"
        name='new-server-test'
        imageRef='%(image_id)s'
        flavorRef='1'
>
        <OS-SCH-HNT:scheduler_hints>
                <hypervisor>xen</hypervisor>
                <near>%(image_near)s</near>
        </OS-SCH-HNT:scheduler_hints>
</server>
