{
    "extensions": [
        {
            "alias": "extensions",
            "description": "Extension information.",
            "name": "extensions",
            "version": 1
        },
        {
            "alias": "flavors",
            "description": "Flavors Extension.",
            "name": "flavors",
            "version": 1
        }
    ]
}
