{
    "versions": [
        {
            "id": "v2.0",
            "links": [
                {
                    "href": "http://openstack.example.com/v2/",
                    "rel": "self"
                }
            ],
            "status": "CURRENT",
            "updated": "2011-01-21T11:33:21Z"
        },
        {
            "id": "v3.0",
            "links": [
                {
                    "href": "http://openstack.example.com/v3/",
                    "rel": "self"
                }
            ],
            "status": "EXPERIMENTAL",
            "updated": "2013-07-23T11:33:21Z"
        }
    ]
}
