{
    "cloudpipes": [
        {
            "created_at": "%(timestamp)s",
            "instance_id": "%(uuid)s",
            "internal_ip": "%(ip)s",
            "project_id": "%(project_id)s",
            "public_ip": "%(ip)s",
            "public_port": 22,
            "state": "down"
        }
    ]
}
