<?xml version="1.0" encoding="UTF-8"?>
  <events>
    <event>
      <name>%(name)s</name>
      <tag>%(tag)s</tag>
      <status>%(status)s</status>
      <server_uuid>%(uuid)s</server_uuid>
    </event>
  </events>
