<?xml version='1.0' encoding='UTF-8'?>
<server xmlns:atom="http://www.w3.org/2005/Atom"
        xmlns="http://docs.openstack.org/compute/api/v1.1"
        id="%(id)s" adminPass="%(password)s">
  <metadata/>
  <atom:link href="%(host)s/v2/openstack/servers/%(uuid)s" rel="self"/>
  <atom:link href="%(host)s/openstack/servers/%(uuid)s" rel="bookmark"/>
</server>
