{
    "agents": [
         {
            "hypervisor": "%(hypervisor)s",
            "os": "%(os)s",
            "architecture": "%(architecture)s",
            "version": "%(version)s",
            "md5hash": "%(md5hash)s",
            "url": "%(url)s",
            "agent_id": "%(agent_id)d"
        }
    ]
}
