{
    "addresses": {
        "private": [
            {
                "addr": "%(ip)s",
                "mac_addr": "aa:bb:cc:dd:ee:ff",
                "type": "fixed",
                "version": 4
            }
        ]
    }
}
