<?xml version='1.0' encoding='UTF-8'?>
<flavors xmlns:OS-FLV-EXT-DATA="http://docs.openstack.org/compute/ext/flavor_extra_data/api/v1.1" xmlns:atom="http://www.w3.org/2005/Atom" xmlns="http://docs.openstack.org/compute/api/v1.1">
  <flavor disk="1" vcpus="1" ram="512" name="m1.tiny" id="1" OS-FLV-EXT-DATA:ephemeral="0">
    <atom:link href="%(host)s/v2/openstack/flavors/1" rel="self"/>
    <atom:link href="%(host)s/openstack/flavors/1" rel="bookmark"/>
  </flavor>
  <flavor disk="20" vcpus="1" ram="2048" name="m1.small" id="2" OS-FLV-EXT-DATA:ephemeral="0">
    <atom:link href="%(host)s/v2/openstack/flavors/2" rel="self"/>
    <atom:link href="%(host)s/openstack/flavors/2" rel="bookmark"/>
  </flavor>
  <flavor disk="40" vcpus="2" ram="4096" name="m1.medium" id="3" OS-FLV-EXT-DATA:ephemeral="0">
    <atom:link href="%(host)s/v2/openstack/flavors/3" rel="self"/>
    <atom:link href="%(host)s/openstack/flavors/3" rel="bookmark"/>
  </flavor>
  <flavor disk="80" vcpus="4" ram="8192" name="m1.large" id="4" OS-FLV-EXT-DATA:ephemeral="0">
    <atom:link href="%(host)s/v2/openstack/flavors/4" rel="self"/>
    <atom:link href="%(host)s/openstack/flavors/4" rel="bookmark"/>
  </flavor>
  <flavor disk="160" vcpus="8" ram="16384" name="m1.xlarge" id="5" OS-FLV-EXT-DATA:ephemeral="0">
    <atom:link href="%(host)s/v2/openstack/flavors/5" rel="self"/>
    <atom:link href="%(host)s/openstack/flavors/5" rel="bookmark"/>
  </flavor>
</flavors>
