<?xml version='1.0' encoding='UTF-8'?>
<cloudpipes>
  <cloudpipe>
    <public_ip>%(ip)s</public_ip>
    <created_at>%(timestamp)s</created_at>
    <public_port>22</public_port>
    <state>down</state>
    <instance_id>%(uuid)s</instance_id>
    <internal_ip>%(ip)s</internal_ip>
    <project_id>%(project_id)s</project_id>
  </cloudpipe>
</cloudpipes>
