{
    "availabilityZoneInfo": [
        {
            "zoneName": "zone-1",
            "zoneState": {
                "available": true
            },
            "hosts": null
        },
        {
            "zoneName": "zone-2",
            "zoneState": {
                "available": false
            },
            "hosts": null
        }
    ]
}