{
    "instance_actions": [
        {
            "action": "%(action)s",
            "instance_uuid": "%(uuid)s",
            "request_id": "%(request_id)s",
            "user_id": "%(integer_id)s",
            "project_id": "%(integer_id)s",
            "start_time": "%(timestamp)s",
            "message": ""
        },
        {
            "action": "%(action)s",
            "instance_uuid": "%(uuid)s",
            "request_id": "%(request_id)s",
            "user_id": "%(integer_id)s",
            "project_id": "%(integer_id)s",
            "start_time": "%(timestamp)s",
            "message": ""
        }
    ]
}
