#!/usr/bin/env python
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module to authenticate Consoles."""

from oslo.config import cfg


consoleauth_topic_opt = cfg.StrOpt('consoleauth_topic',
        default='consoleauth',
        help='The topic console auth proxy nodes listen on')

CONF = cfg.CONF
CONF.register_opt(consoleauth_topic_opt)
