<?xml version="1.0" encoding="UTF-8"?>
<createImage
    xmlns="http://docs.openstack.org/compute/api/v1.1"
    name="%(name)s">
    <metadata>
        <meta key="%(meta_var)s">%(meta_val)s</meta>
    </metadata>
</createImage>
