{
    "aggregate": {
        "availability_zone": "nova",
        "created_at": "%(timestamp)s",
        "deleted": 0,
        "deleted_at": null,
        "id": %(aggregate_id)s,
        "name": "name",
        "updated_at": null
    }
}
