<?xml version="1.0" encoding="UTF-8"?>
  <snapshot>
    <display_name>%(snapshot_name)s</display_name>
    <display_description>%(description)s</display_description>
    <volume_id>%(volume_id)s</volume_id>
    <force>false</force>
  </snapshot>
