<?xml version='1.0' encoding='UTF-8'?>
<floating_ip_info>
  <item>
    <interface>eth0</interface>
    <instance_uuid>None</instance_uuid>
    <project_id>None</project_id>
    <pool>nova</pool>
    <address>10.10.10.1</address>
  </item>
  <item>
    <interface>eth0</interface>
    <instance_uuid>None</instance_uuid>
    <project_id>None</project_id>
    <pool>nova</pool>
    <address>10.10.10.2</address>
  </item>
  <item>
    <interface>eth0</interface>
    <instance_uuid>None</instance_uuid>
    <project_id>None</project_id>
    <pool>nova</pool>
    <address>10.10.10.3</address>
  </item>
</floating_ip_info>