{
    "network":
        {
            "bridge": "br100",
            "bridge_interface": "eth0",
            "broadcast": "%(ip)s",
            "cidr": "10.0.0.0/29",
            "cidr_v6": null,
            "created_at": "%(timestamp)s",
            "deleted": false,
            "deleted_at": null,
            "dhcp_start": "%(ip)s",
            "dns1": null,
            "dns2": null,
            "gateway": "%(ip)s",
            "gateway_v6": null,
            "host": "nsokolov-desktop",
            "id": "%(id)s",
            "injected": false,
            "label": "mynet_0",
            "multi_host": false,
            "netmask": "%(ip)s",
            "netmask_v6": null,
            "priority": null,
            "project_id": "1234",
            "rxtx_base": null,
            "updated_at": "%(timestamp)s",
            "vlan": 100,
            "vpn_private_address": "%(ip)s",
            "vpn_public_address": "%(ip)s",
            "vpn_public_port": 1000
        }
}
