<?xml version='1.0' encoding='UTF-8'?>
<server_group xmlns:atom="http://www.w3.org/2005/Atom" xmlns="http://docs.openstack.org/compute/api/v1.1" id="%(id)s" name="test">
  <policies>
  <policy>anti-affinity</policy>
  </policies>
  <members/>
  <metadata/>
</server_group>
