{
    "nodes": [
        {
            "cpus": 8,
            "id": %(node_id)s,
            "instance_uuid": null,
            "interfaces": [
                {
                    "address": "%(address)s",
                    "datapath_id": null,
                    "id": %(interface_id)s,
                    "port_no": null
                }
            ],
            "local_gb": 128,
            "memory_mb": 8192,
            "pm_address": "10.1.2.3",
            "pm_user": "pm_user",
            "pxe_config_path": null,
            "service_host": "host",
            "task_state": null,
            "terminal_port": 8000,
            "updated_at": null,
            "uuid": "%(node_uuid)s"
        }
    ]
}
