<?xml version='1.0' encoding='UTF-8'?>
<versions xmlns:atom="http://www.w3.org/2005/Atom" xmlns="http://docs.openstack.org/common/api/v1.0">
  <version status="CURRENT" updated="2011-01-21T11:33:21Z" id="v2.0">
    <atom:link href="http://openstack.example.com/v2/" rel="self"/>
  </version>
  <version status="EXPERIMENTAL" updated="2013-07-23T11:33:21Z" id="v3.0">
    <atom:link href="http://openstack.example.com/v3/" rel="self"/>
  </version>
</versions>
