<?xml version='1.0' encoding='UTF-8'?>
<virtual_interfaces
        xmlns:OS-EXT-VIF-NET="http://docs.openstack.org/compute/ext/extended-virtual-interfaces-net/api/v1.1"
        xmlns="http://docs.openstack.org/compute/api/v1.1">
  <virtual_interface id="%(id)s"
                     mac_address="%(mac_addr)s"
                     OS-EXT-VIF-NET:net_id="%(id)s"/>
</virtual_interfaces>
