{
    "flavor_access": [
        {
            "flavor_id": "%(flavor_id)s",
            "tenant_id": "fake_tenant"
        },
        {
            "flavor_id": "%(flavor_id)s",
            "tenant_id": "openstack"
        }
    ]
}
